from category_encoders import OrdinalEncoder
import numpy as np
import pandas as pd

from mondobrain.datautils import apply_rule
from mondobrain.utils.data import is_continuous

from ._target import target_metrics


def score(
    df: pd.DataFrame,
    outcome: str = None,
    target: str = None,
    rule: dict = None,
    population: pd.DataFrame = None,
):
    """Score a dataframe

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to score. If rule is set, then df should represent the original
        population. If population is set, then df should represent the sample to score.

    outcome : str, default=None
        Which column to use as the outcome for the score (sometimes referred to as a
        target feature). If None, then the first column is selected

    target: str, default=None
        The class in the outcome column to target. For a continuous outcome this should
        either be `min` or `max`. If None, then the first class or `max` is selected.

    rule: dict, default=None
        The solver rule to use for scoring. This or population must be set, but not
        both. If set, df should be the population dataset

    population: pd.DataFrame, default=None
        The population dataset to base the score off of. If set, df should be the sample
        dataset.

    Returns
    -------
    score: float
        The score of the sample and population
    """
    # Quick checks to make sure that one of rule and population is set.
    if rule is None and population is None:
        raise ValueError("one of `rule` or `population` must be set")

    if rule is not None and population is not None:
        raise ValueError("only one of `rule` or `population` may be set")

    if outcome is None:
        outcome = df.columns[0]

    if target is None:
        target = df[outcome].iloc[0] if df[outcome].dtype == np.object else "max"

    if population is not None:
        sample = df

    if rule is not None:
        population = df
        sample = apply_rule(df, rule)

    # Encode our outcome column if the outcome column is continuous
    if is_continuous(population[outcome].dtype):
        enc = OrdinalEncoder(cols=[outcome], handle_missing="return_nan")
        population = enc.fit_transform(population)
        sample = enc.transform(sample)

    # Get metrics for population and sample
    pop_metrics = target_metrics(population[outcome], target)
    smp_metrics = target_metrics(sample[outcome], target)

    # Return 0 if we our size is 0
    if smp_metrics["size"] == 0 or pop_metrics["size"] == 0:
        return 0

    return (
        np.sqrt(smp_metrics["size"])
        * (smp_metrics["mean"] - pop_metrics["mean"])
        / pop_metrics["std"]
    )



from typing import Dict, List, Union

from mondobrain.core.meta import Direction, Inverse

from .requestor import APIRequestor


def meta_class_start(
    outcome: str = None,
    timeout: int = None,
    min_size_frac: float = None,
    # case_points: dict = None, DISABLED FOR NOW
    classes: List[str] = None,
    direction: Direction = Direction.both,
    encode: List[str] = None,
    inverse: Inverse = None,
    depth: int = None,
    parallel: bool = False,
    data: List[Dict[str, Union[None, str, int, float]]] = None,
    **kwargs
):
    """
    Start a MetaMondo exploration using a class-based strategy.

    Feature rules are gathered by learning on multiple classes in multiple directions.
    For discrete outcome variables, this means in or out of each outcome modality and
    for continuous outcome variables, this means min and/or max

    Parameters
    ----------
    classes: List[str] (optional), default=None
        For discrete outcomes, a list of modalities to explore.
        For continuous outcomes, a list containing 'min' and/or 'max'.
        If None, explore all classes.

    direction: 'in', 'out', 'both' (optional), default='both'
        Direction to explore each outcome class.  Only applicable for discrete
        outcomes.

        * **in**: find feature rules inside of the outcome class
        * **out**: find feature rules outside of the outcome class
        * **both**: do both in and out
        * **None**: do not explore directions (for continuous outcome variables)

    inverse: 'in_rule' (optional), default=None
        Strategy for finding "inverse" rules when a feature rule is found

        * **in_rule**: find a rule inside the feature rule in the opposite direction
        * **None**: do not add inverse rules

    depth: int (optional), default=1
        The number of exploration cycles.

    parallel: bool (optional), default=False
        Explore each outcome class independently and in parallel.

    outcome: str
        See **api.solve_start**

    data: List[dict]
        See **api.solve_start**

    timeout: int (optional)
        See **api.solve_start**

    min_size_frac: float (optional)
        See **api.solve_start**
    """
    params = {
        "outcome": outcome,
        "timeout": timeout,
        "min_size_frac": min_size_frac,
        "classes": classes,
        "direction": direction,
        "encode": encode,
        "inverse": inverse,
        "depth": depth,
        "parallel": parallel,
        "data": data,
    }

    requestor = APIRequestor(**kwargs)
    return requestor.request("post", "meta.class.start", params=params)


def meta_exclusion_set_start(
    outcome: str = None,
    timeout: int = None,
    min_size_frac: float = None,
    # case_points: dict = None, DISABLED FOR NOW
    min_size: int = None,
    classes: List[str] = None,
    direction: Direction = Direction.both,
    encode: List[str] = None,
    depth: int = None,
    parallel: bool = False,
    data: List[Dict[str, Union[None, str, int, float]]] = None,
    **kwargs
):
    """
    Start a MetaMondo exploration using an exclusion set strategy.

    Feature rules are gathered by learning a rule, then learning again on the exclusion
    set (the set of data points outside of the rule), repeating until the exclusion set
    reaches a minimum number of data points.

    Parameters
    ----------
    min_size: int (optional), default=10
        The minimum size of the exclusion set.  Stops exploration when the
        exclusion set is below this size.

    classes: List[str] (optional), default=None
        For discrete outcomes, a list of modalities to explore.
        For continuous outcomes, a list containing 'min' and/or 'max'.
        If None, explore all classes.

    direction: 'in', 'out', 'both' (optional), default='both'
        Direction to explore each outcome class.  Only applicable for discrete
        outcomes.

        * **in**: find feature rules inside of the outcome class
        * **out**: find feature rules outside of the outcome class
        * **both**: do both in and out
        * **None**: do not explore directions (for continuous outcome variables)

    depth: int (optional), default=1
        The number of exploration cycles.

    parallel: bool (optional), default=False
        Explore each outcome class independently and in parallel.

    outcome: str
        See **api.solve_start**

    data: List[dict]
        See **api.solve_start**

    timeout: int (optional)
        See **api.solve_start**

    min_size_frac: float (optional)
        See **api.solve_start**
    """
    params = {
        "outcome": outcome,
        "timeout": timeout,
        "min_size_frac": min_size_frac,
        "min_size": min_size,
        "classes": classes,
        "direction": direction,
        "encode": encode,
        "depth": depth,
        "parallel": parallel,
        "data": data,
    }

    requestor = APIRequestor(**kwargs)
    return requestor.request("post", "meta.exclusionset.start", params=params)


def meta_result(id: str, **kwargs):
    """Read meta solve results (rules) from the API

    Parameters
    ----------
    id: str
        The task ID as returned from **api.meta_class_start** or
        **api.meta_exclusion_set_start**
    """
    params = {"id": id}

    requestor = APIRequestor(**kwargs)

    return requestor.request("post", "meta.result", params=params)



from typing import List, Union

import numpy as np
import pandas as pd
from sklearn.metrics import (
    accuracy_score,
    log_loss,
    mean_absolute_error,
    mean_squared_error,
    precision_score,
    recall_score,
)
from sklearn.model_selection import KFold

from mondobrain.core.frame import MondoDataFrame
from mondobrain.core.series import MondoSeries
from mondobrain.dd_transformer import DDTransformer

from .constants import MIN_SOLVER_SIZE, RULE_PREFIX
from .data import is_discrete


def has_enough_values(df, target_class, target_variable):
    col = df[target_variable]

    if is_discrete(col.dtype):
        size = col[col == target_class].shape[0]
    else:
        if col.std() == 0:
            return False
        size = col.shape[0]

    return size >= MIN_SOLVER_SIZE


def one_hot_encode(df: MondoDataFrame, cols: List[str]) -> MondoDataFrame:
    if cols is None:
        cols = []
    cols = [c for c in cols if len(df[c].unique()) <= 12]

    if cols:
        return MondoDataFrame(pd.get_dummies(df, columns=cols))
    else:
        return df


def encode_column(col: MondoSeries, encoder: DDTransformer) -> MondoSeries:
    encoded, encoded_col_label = encoder.original_to_encoded(col.name, col)
    ms_encoded = MondoSeries(encoded, name=encoded_col_label)

    # Need to change encoded nulls back to nulls
    ms_encoded.where(~col.reset_index(drop=True).isnull(), np.nan, inplace=True)

    return ms_encoded


def encode_value(
    mdf: MondoDataFrame, outcome: str, value: any, encoder: DDTransformer
) -> str:
    if not np.issubdtype(mdf[outcome].dtype, np.number):
        value_encoded = encode_column(MondoSeries([value], name=outcome), encoder)[0]
    else:
        value_encoded = value

    return value_encoded


def encode_dataframe(mdf: MondoDataFrame, encoder: DDTransformer) -> MondoDataFrame:
    """
    DOCSTRING
    :return: encoded MondoDataFrame,
    """
    encoded_dict = dict()

    for column, values in mdf.iteritems():
        encoded_col = encode_column(values, encoder)
        encoded_dict[encoded_col.name] = encoded_col

    encoded_dataframe = MondoDataFrame.from_dict(encoded_dict)

    return encoded_dataframe


def decode_column(col: MondoSeries, encoder: DDTransformer) -> MondoSeries:
    numeric_col_label = encoder._column_decoder[col.name]
    ms_decoded = encoder.encoded_to_original(numeric_col_label, col)

    return ms_decoded


def decode_value(col: MondoSeries, value: any, encoder: DDTransformer) -> MondoSeries:
    numeric_col_label = encoder._column_decoder[col.name]
    ms_decoded = encoder.encoded_to_original(numeric_col_label, np.array([value]))

    return ms_decoded[0]


def decode_dataframe(mdf: MondoDataFrame, encoder: DDTransformer):
    decoded_dict = dict()
    for col, ser in mdf.iteritems():
        decoded_dict[col] = decode_column(ser, encoder)

    return MondoDataFrame.from_dict(decoded_dict)


def sample_if_needed(df, target_class, target_variable, random_state=1337):

    SAMPLING_SIZE = 2500
    SAMPLING_CAP = 3500

    if df.shape[0] > SAMPLING_CAP:

        was_sampled = True

        # if the outcome is not numeric, we might use oversampled outcomes and
        # the total number of points are less than 0.15 of the overall
        if not np.issubdtype(df[target_variable], np.number):
            # uniform stratified sampling across the outcome column
            # if no target class is specified
            if not target_class:
                df = (
                    df.groupby(target_variable, group_keys=False)
                    .apply(
                        lambda x: x.sample(
                            int(np.rint(SAMPLING_SIZE * x.shape[0] / df.shape[0]))
                        )
                    )
                    .sample(frac=1)
                    .reset_index(drop=True)
                )
            elif df[target_variable].value_counts()[target_class] / df.shape[0] < 0.15:
                # If they are, we will over sample them
                df = oversampled_result(
                    df, target_variable, target_class, SAMPLING_SIZE
                )
            elif df[target_variable].value_counts().shape[0] == 2:
                target_classes = df[target_variable].value_counts().index
                alternate_class = target_classes[
                    ~np.in1d(target_classes, target_class)
                ][0]
                if (
                    df[target_variable].value_counts()[alternate_class] / df.shape[0]
                    < 0.15
                ):
                    # If they are, we will over sample them
                    df = oversampled_result(
                        df, target_variable, alternate_class, SAMPLING_SIZE
                    )
                else:
                    df = df.sample(
                        SAMPLING_SIZE, random_state=random_state
                    ).reset_index(drop=True)
            else:
                df = df.sample(SAMPLING_SIZE, random_state=random_state).reset_index(
                    drop=True
                )
        else:
            df = df.sample(SAMPLING_SIZE, random_state=random_state).reset_index(
                drop=True
            )

    else:
        was_sampled = False

    return df, was_sampled


def oversampled_result(df, outcome, target_class, sampling_size):
    target_prob = 1 - (df[outcome] == target_class).sum() / df.shape[0]
    opposing_prob = 1 - target_prob
    pick_probabilities = np.where(
        df[outcome] == target_class, target_prob, opposing_prob
    )
    pick_probabilities = pick_probabilities / pick_probabilities.sum()
    # this sampling approach will choose every value of the underrepresented
    # class until they are all chosen
    df_sample = df.sample(
        sampling_size, random_state=1337, weights=pick_probabilities
    ).reset_index(drop=True)
    return df_sample


def decode_rule(rule: dict, encoder: DDTransformer) -> dict:
    rule_decoded = dict()

    for key, condition in rule.items():
        if key.startswith(RULE_PREFIX):
            # MetaMondo condition
            key_decoded = key
            condition_decoded = condition["lo"]
        elif isinstance(condition, str):
            # Discrete condition
            col_label = encoder._column_decoder[key]
            values_encoded = MondoSeries([condition])
            values_decoded = encoder.encoded_to_original(col_label, values_encoded)

            key_decoded = values_decoded.name
            condition_decoded = values_decoded[0]
        else:
            # Continuous condition
            col_label = encoder._column_decoder[key]
            values_encoded = MondoSeries(list(condition.values()))
            values_decoded = encoder.encoded_to_original(col_label, values_encoded)
            bounds_decoded = values_decoded.values.tolist()

            key_decoded = values_decoded.name
            condition_decoded = {"lo": bounds_decoded[0], "hi": bounds_decoded[1]}

        rule_decoded[key_decoded] = condition_decoded

    return rule_decoded


def encode_rule(rule: dict, encoder: DDTransformer) -> dict:
    rule_encoded = dict()

    for key, condition in rule.items():
        key_encoded = encoder._column_encoder[key]

        if isinstance(condition, str):
            # Discrete condition
            values = MondoSeries([condition])
            values_encoded = encoder.original_to_encoded(key, values)

            key_encoded = values_encoded.name
            condition_encoded = values_encoded[0]
        else:
            # Continuous condition
            values = MondoSeries(list(condition.values()))
            values_encoded = encoder.original_to_encoded(key, values)

            bounds_encoded = values_encoded.values.tolist()

            key_encoded = values_encoded.name
            condition_encoded = {"lo": bounds_encoded[0], "hi": bounds_encoded[1]}

        rule_encoded[key_encoded] = condition_encoded


def apply_rule_to_df(df: pd.DataFrame, rule: dict):
    """
    fn that enables applying a rule learned on another database
    onto a database with the same column names

    rule: format {variable: {'low': low_value, 'high': high_value}}
    """

    def get_mask(df: pd.DataFrame, key: str, condition: Union[str, dict]):
        if isinstance(condition, dict):
            # Continuous condition
            try:
                lo = condition["lo"]
                hi = condition["hi"]

                mask = (df[key] >= lo) & (df[key] <= hi)
            except TypeError:
                raise TypeError(
                    """Attempted to apply a numerical condition to a
                        non-numerical dataframe column"""
                )
        else:
            # Discrete condition
            mask = df[key] == condition

        return mask

    if not rule:
        return df.drop(df.index)

    masks = np.all(
        np.array([get_mask(df, key, cond) for key, cond in rule.items()]), axis=0
    )
    return df[masks]


def get_stats(df: pd.DataFrame, outcome, target) -> dict:
    if df[outcome].dtype == np.object:
        values = (df[outcome] == target).astype(np.int)
    else:
        values = df[outcome]
        if target == "min":
            values *= -1

    stats = {
        "mean": values.mean(),
        "std": values.std(ddof=0),
        "size": int(values.size),
        "size_above_mean": 0,
    }

    if values.mean() > 0:
        stats["size_above_mean"] = int((values >= values.mean()).sum())

    return stats


def calculate_score(sample_stats, pop_stats):
    if sample_stats["size"] == 0 or pop_stats["size"] == 0:
        score = 0
    else:
        score = (
            np.sqrt(sample_stats["size"])
            * (sample_stats["mean"] - pop_stats["mean"])
            / pop_stats["std"]
        )

    return score


def score_rule(dataset, rule, outcome: str, target: str):
    pop_stats = get_stats(dataset, outcome, target)
    df_rule = apply_rule_to_df(dataset, rule)
    sample_stats = get_stats(df_rule, outcome, target)

    return calculate_score(sample_stats, pop_stats)


def cross_validate_rule(rule, m_X, m_y, cv=3, score=None):
    if isinstance(cv, int) or isinstance(cv, float):
        cv = KFold(n_splits=cv, shuffle=True)

    val_scores = []
    val_sizes = []

    for val_index, _ in cv.split(m_X, m_y):
        X_val = m_X.iloc[val_index]
        y_val = m_y.iloc[val_index]
        mdf_val = MondoDataFrame(pd.concat((MondoDataFrame(y_val), X_val), axis=1))

        outcome = m_y.name
        target = m_y.target_class

        df_applied = apply_rule_to_df(mdf_val, rule)
        sample_stats = get_stats(df_applied, outcome, target)
        val_score = score_rule(mdf_val, rule, outcome, target)
        val_size = sample_stats["size"]

        val_scores.append(val_score)
        val_sizes.append(val_size)

    if score is not None:
        mse = np.mean((np.array(val_scores) - np.array(score)) ** 2)
    else:
        mse = None

    validation = {
        "scores": val_scores,
        "sizes": val_sizes,
        "avg_loss": mse,
    }

    return validation


def prettify_rule(ugly_rule, dataset):
    pretty_rule = dict()
    for col, bounds in ugly_rule.items():
        if is_discrete(dataset[col].dtype):
            pretty_rule[col] = {"discrete": bounds["low"]}
        else:
            pretty_rule[col] = bounds

    return pretty_rule


def model_report(y_true, y_pred):
    model_dict = {}
    # MAE and MSE are for numeric (float or int) loss functions
    if np.issubdtype(y_true.dtype, np.float64) or np.issubdtype(y_true.dtype, np.int64):
        try:
            model_dict["mean_abs_error"] = mean_absolute_error(y_true, y_pred)
        except Exception:
            model_dict["mean_abs_error"] = None
        try:
            model_dict["mean_sq_error"] = mean_squared_error(y_true, y_pred)
        except Exception:
            model_dict["mean_sq_error"] = None
    # log_loss is more for binary (either 0,1 or True, False (boolean))
    elif np.issubdtype(y_true.dtype, np.bool):
        try:
            model_dict["log_loss"] = log_loss(y_true, y_pred)
        except Exception:
            model_dict["log_loss"] = None
        try:
            model_dict["accuracy_score"] = accuracy_score(y_true, y_pred)
        except Exception:
            model_dict["accuracy_score"] = None
        try:
            model_dict["precision_score"] = precision_score(y_true, y_pred)
        except Exception:
            model_dict["precision_score"] = None
        try:
            model_dict["recall_score"] = recall_score(y_true, y_pred)
        except Exception:
            model_dict["recall_score"] = None
    # labeled or discrete data
    else:
        try:
            model_dict["precision_score"] = precision_score(
                y_true, y_pred, average=None
            )
        except Exception:
            model_dict["precision_score"] = None
        try:
            model_dict["recall_score"] = recall_score(y_true, y_pred, average=None)
        except Exception:
            model_dict["recall_score"] = None
    return model_dict



from typing import Dict, List

import pandas as pd

from mondobrain.core.frame import MondoDataFrame
from mondobrain.dd_transformer import DDTransformer
from mondobrain.utils import utilities

from .constants import RULE_PREFIX
from .data import is_discrete


def get_meta_name(meta: dict, with_stats: bool = False) -> str:
    class_lbl = "[{0}]".format(meta["class_"])
    dir_lbl = "[dir={0}]".format(meta["direction"]) if meta["direction"] else ""
    inv_lbl = "[inv={0}]".format(meta["inverse"]) if meta["inverse"] else ""
    depth_lbl = "[d={0}]".format(meta["depth"]) if meta["depth"] else ""
    iter_lbl = "[i={0}]".format(meta["iteration"]) if meta["iteration"] else ""
    score_lbl = "[z={:.3f}]".format(float(meta["score"])) if with_stats else ""
    size_lbl = "[n={:.0f}]".format(float(meta["size"])) if with_stats else ""

    return "{0}{1}{2}{3}{4}{5}{6}{7}".format(
        RULE_PREFIX,
        class_lbl,
        dir_lbl,
        iter_lbl,
        inv_lbl,
        depth_lbl,
        score_lbl,
        size_lbl,
    )


def update_meta(rule: dict, meta: dict = None):
    if not rule["meta"]:
        rule["meta"] = {}
    if meta:
        rule["meta"].update(meta)

    rule["meta"]["name"] = get_meta_name(rule["meta"])


def decode_meta_rules(
    rules: List[dict], mdf: MondoDataFrame, outcome: str, encoder: DDTransformer,
) -> List[dict]:
    decoded_rules = []
    meta_rule_name_map = {}

    if is_discrete(mdf[outcome].dtype):
        for meta_rule in rules:
            old_name = meta_rule["meta"]["name"]

            meta_decoded = {
                "class_": utilities.decode_value(
                    mdf[outcome], meta_rule["meta"]["class_"], encoder
                )
            }
            update_meta(meta_rule, meta_decoded)

            new_name = meta_rule["meta"]["name"]
            meta_rule_name_map[old_name] = new_name

        for meta_rule in rules:
            rule_decoded = utilities.decode_rule(meta_rule["rule"], encoder)

            old_keys = [
                key for key, cond in rule_decoded.items() if key.startswith(RULE_PREFIX)
            ]
            for key in old_keys:
                new_key = meta_rule_name_map[key]
                rule_decoded[new_key] = rule_decoded.pop(key)

            meta_rule_decoded = {
                "rule": rule_decoded,
                "meta": meta_rule["meta"],
            }

            decoded_rules.append(meta_rule_decoded)
    else:
        for meta_rule in rules:
            rule_decoded = utilities.decode_rule(meta_rule["rule"], encoder)
            meta_rule_decoded = {
                "rule": rule_decoded,
                "meta": meta_rule["meta"],
            }

            decoded_rules.append(meta_rule_decoded)

    return decoded_rules


def print_rule_trees(rules: List[dict], class_: str = None, depth: int = None):
    rule_map = __get_rule_map(rules)

    filt_rules = rules
    if class_:
        filt_rules = [x for x in filt_rules if class_ == x["meta"]["class_"]]
    if depth:
        filt_rules = [x for x in filt_rules if depth == x["meta"]["depth"]]

    for rule in filt_rules:
        __print_rule_tree_helper(rule, rule_map, is_root=True)


def apply_meta_rules_to_df(rules: List[dict], df: pd.DataFrame):
    filt_rules = [x for x in rules if x["meta"]["name"] not in df.columns]
    rule_map = __get_rule_map(filt_rules)
    curr_df = df.copy()

    feat_rule_map = {}

    for rule in filt_rules:
        feat_rule_map[rule["meta"]["name"]] = []

        for key, cond in rule["rule"].items():
            if key.startswith(RULE_PREFIX):
                feat_rule_map[rule["meta"]["name"]].append(key)

    while feat_rule_map:
        applied_rules = []

        for name, feat_rules in feat_rule_map.items():
            if len(feat_rules) == 0:
                rule = rule_map[name]
                df_rule = utilities.apply_rule_to_df(curr_df, rule["rule"])

                curr_df[name] = curr_df.index.isin(df_rule.index).astype(int)
                applied_rules.append(name)

        if len(applied_rules) == 0:
            break

        for name in applied_rules:
            del feat_rule_map[name]

        for name, feat_rules in feat_rule_map.items():
            feat_rule_map[name] = [x for x in feat_rules if x not in applied_rules]

    return curr_df


def __get_rule_map(rules: List[dict]) -> Dict[str, dict]:
    return {x["meta"]["name"]: x for x in rules}


def __print_rule_tree_helper(
    rule: dict, rule_map: Dict[str, dict], prefix: str = "", is_root: bool = False,
):

    TREE_SPACE, TREE_BRANCH, TREE_TEE, TREE_END = "    ", "│   ", "├── ", "└── "

    filters = rule["rule"]

    if is_root:
        print(get_meta_name(rule["meta"], with_stats=True))

    pointers = [TREE_TEE] * (len(filters) - 1) + [TREE_END]

    for pointer, col in zip(pointers, filters):
        filt = filters[col]
        col_lbl = prefix + pointer + col + ":"

        if type(filt) is dict and "lo" in filt and "hi" in filt:
            if filt["lo"] == filt["hi"]:
                print(col_lbl, filt["lo"])
            else:
                print(col_lbl, "[{0}, {1}]".format(filt["lo"], filt["hi"]))
        else:
            print(col_lbl, filt)

        if col.startswith(RULE_PREFIX):
            extension = TREE_BRANCH if pointer == TREE_TEE else TREE_SPACE

            __print_rule_tree_helper(rule_map[col], rule_map, prefix=prefix + extension)

    if is_root:
        print()



from io import BytesIO
import json

import numpy as np
import pandas as pd
from sklearn.model_selection import KFold, train_test_split

import mondobrain as mb
from mondobrain.core.frame import MondoDataFrame
from mondobrain.core.meta import BaseExploration
from mondobrain.core.series import MondoSeries
from mondobrain.dd_transformer import DDTransformer
from mondobrain.error import NotEnoughPointsError
from mondobrain.utils import meta as meta_utils, utilities


class Solver:
    """
    MondoBrain Solver object.

    Applies a stochastic search for the global maximum Z-score with respect to
    a defined dependent variable and target class, in cases of "discrete" variables
    or target min/max mean in the cases of "continuous" variables.

    Parameters
    ----------
    min_size_frac: float (optional), default=0.2
        Value between 0.0 and 1.0 that defines the minimum number of points needed
        for a valid rule discovered by the MondoBrain solver.

    min_purity: float (optional), default=0.0
        Value between 0.0 and 1.0 that defines the minimum purity needed for a valid
        rule discovered by the MondoBrain solver.

        Purity here is defined as the mean of the target variable distribution.

    max_cycles: int (optional), default=90
        Value greater than 0 that defines the total cycles the MondoBrain solver should
        commit in order to find a valid rule.

    Examples
    --------
    Using the solver to find a rule

    >>> mb.api_key = "<API KEY>"
    >>> mb.api_secret = "<API SECRET>"
    >>> solver = mb.Solver()
    >>> solver.fit(mdf_explorable, mdf_outcome)
    >>> solver.rule
    {
        'sex': {'low': 'male', 'high': 'male'},
        'class': {'low': 2, 'high': 3},
        'parch': {'low': 0, 'high': 0},
        'ticketnumber': {'low': 2152.0, 'high': 3101281.0}
    }
    >>> solver.score
    12.974682681486312


    """

    def __init__(
        self, min_size_frac=0.2, min_purity=0.0, max_cycles=90,
    ):
        self.min_size_frac = min_size_frac
        self.min_purity = min_purity
        self.max_cycles = max_cycles

    @staticmethod
    def _is_one_dim_solve(explorable_vars):
        return explorable_vars.shape[1] == 1

    @staticmethod
    def _get_timeout(n_points, n_features, lg_coefficient=3.5, max_cycles=90):

        lg_timeout = int(lg_coefficient * np.sqrt(n_points * n_features / 2))
        min_timeout = np.minimum(lg_timeout, max_cycles)

        return int(np.maximum(1, min_timeout))

    @staticmethod
    def _get_dataset_buffer(df: pd.DataFrame):
        buffer = BytesIO()
        df.to_parquet(buffer)
        buffer.seek(0)
        return buffer

    def _run_solve(self, mdf: MondoDataFrame, outcome: str, target: str):
        encoder = DDTransformer(mdf)
        edf = utilities.encode_dataframe(mdf, encoder)
        target_encoded = utilities.encode_value(mdf, outcome, target, encoder)
        outcome_encoded = encoder.original_to_encoded_column(outcome)

        edf, _ = utilities.sample_if_needed(edf, target_encoded, outcome_encoded)

        enough_points = utilities.has_enough_values(
            edf, target_encoded, outcome_encoded
        )
        if not enough_points:
            raise NotEnoughPointsError(
                "Cannot run solver unless there are at least {0} points.".format(
                    utilities.MIN_SOLVER_SIZE
                )
            )

        # Solve Behavior isn't supported by file endpoint yet
        timeout = self._get_timeout(
            edf.shape[0], edf.shape[1] - 1, max_cycles=self.max_cycles
        )

        buf = Solver._get_dataset_buffer(edf)
        task = mb.api.solve_start_file(
            outcome=outcome_encoded,
            target=target_encoded,
            data=buf,
            timeout=timeout,
            min_size_frac=self.min_size_frac,
        )
        result = mb.api.solve_result(id=task["id"])

        rule_encoded = result["rule"]
        rule_decoded = utilities.decode_rule(rule_encoded, encoder)

        # Stats
        if mdf[outcome].var_type == "discrete":
            # Discrete Stats - operate in original space
            mdf_applied = utilities.apply_rule_to_df(mdf, rule_decoded)

            sample_stats = utilities.get_stats(mdf_applied, outcome, target)
            score = utilities.score_rule(mdf, rule_decoded, outcome, target)
        else:
            # Continous Stats - operate in encoded space
            edf_applied = utilities.apply_rule_to_df(edf, rule_encoded)

            sample_stats = utilities.get_stats(
                edf_applied, outcome_encoded, target_encoded
            )
            score = utilities.score_rule(
                edf, rule_encoded, outcome_encoded, target_encoded
            )

        return rule_decoded, sample_stats, score

    @property
    def max_cycles(self):
        return self._max_cycles

    @max_cycles.setter
    def max_cycles(self, value):
        if value > 0:
            self._max_cycles = value
        else:
            raise ValueError("'max_cycles' must be greater than 0.")

    def get_rule_data(self, dataset=None, xrule=False):
        """
        Return a MondoDataFrame that is filtered or not filtered by a rule

        Parameters
        ----------
        dataset: MondoDataFrame of shape (n_samples, n_features)
            Where n_samples is the number of samples and n_features is the number of
            features.

        xrule: Bool, optional (default=False)
            Where `xrule` refers to `not rule` and, if True, returns the portion of the
            dataset not defined by the rule and False returns the portion of the dataset
            defined by the rule.

        """

        if xrule:
            idx = utilities.apply_rule_to_df(dataset, self.rule).index
            xidx = dataset.index[~np.in1d(dataset.index, idx)]
            return dataset.loc[xidx]
        else:
            return utilities.apply_rule_to_df(dataset, self.rule)

    def fit(self, m_X, m_y, cv=None):
        """
        Fit the Solver with the provided data.

        Parameters
        ----------
        m_X: MondoDataFrame of shape (n_samples, n_features)
            Where n_samples is the number of samples and n_features is the number
            of features

        m_y: MondoSeries of shape (n_samples,)
            Where n_samples is the number of samples and aligns with n_samples of m_X

        cv: int or sklearn.model_selection object, optional (default=None)
            The number of folds (n_splits) used in cross validation, or a Sci-Kit
            Learn model splitter, such as StratifiedKFold, ShuffleSplit, etc.  If a
            number is given, the KFold technique is used.  If the value is None,
            the rule does not get cross-validated.

        """

        explorable_vars = [col for col in m_X.columns if col != m_y.name]
        m_X = m_X[explorable_vars].copy()

        if isinstance(cv, int) or isinstance(cv, float):
            cv = KFold(n_splits=cv, shuffle=True)

        if cv is None:
            X_train, X_test, y_train, y_test = m_X, None, m_y, None
        else:
            X_train, X_test, y_train, y_test = train_test_split(
                m_X, m_y, test_size=0.33, random_state=1337
            )
            y_train.target_class = y_test.target_class = m_y.target_class

        mdf = MondoDataFrame(pd.concat((MondoDataFrame(y_train), X_train), axis=1))

        outcome = m_y.name
        target = m_y.target_class

        self.rule, sample_stats, self.score = self._run_solve(mdf, outcome, target)

        self.size = sample_stats["size"]
        self.mean = sample_stats["mean"]
        self.rule_data = self.get_rule_data(mdf)

        if cv is not None:
            self.validation = utilities.cross_validate_rule(
                self.rule, X_test, y_test, cv=cv, score=self.score
            )

    def fit_predicted(
        self, m_X, m_y, m_y_predicted, predicted=False, max_depth=None, min_score=0,
    ):
        """
        Fit the Solver with the provided data but optimizing on the error between
        ground truth and the return values of any statistical model.

        Parameters
        ----------

        m_X: MondoDataFrame of shape (n_samples, n_features)
            Where n_samples is the number of samples and n_features is the number of
            features.

        m_y: MondoSeries of shape (n_samples,)
            Where n_samples is the number of samples and aligns with n_samples of m_X.

        m_y_predicted: MondoSeries, Series, or array (n_samples,), default=None
            Where n_samples is the number of samples and aligns with n_samples of m_y.
            * If provided, `m_y_predicted` will be compare against `m_y` to find a rule
            that explains differences between "correct" and "incorrect" predictions.

        predicted: Bool, optional (default=False)
            Defines the focus of the exploration. To find rules where the provided
            `m_y_predicted` is predicted or where it is not predicted.

        max_depth: Int, optional (default=None)
            Sets a maximum number of levels in to split rules and incidentally the
            subsamples defined by the rules.
            Default `None` allows maximum number of splits.

        min_score: Int, optional (default=0)
            Sets a minimum Z-score found by rules as a threshold for rejecting rules.
            If a rule is rejected, exploration down that branch will bet terminated.

        """

        explorable_vars = [col for col in m_X.columns if col != m_y.name]
        m_X = m_X[explorable_vars].copy()

        if m_y.var_type == "discrete":
            cond = m_y_predicted == m_y
            y_error = MondoSeries(
                np.where(cond, "correct", "incorrect"),
                name="prediction",
                index=m_y.index,
            )
            y_error.target_class = "correct" if predicted else "incorrect"
        else:
            y_error = MondoSeries(
                np.abs(m_y_predicted - m_y), name="residual", index=m_y.index
            )
            y_error.target_class = "min" if predicted else "max"

        mdf = MondoDataFrame(pd.concat((MondoDataFrame(y_error), m_X), axis=1))

        outcome = y_error.name
        target = y_error.target_class

        self.rule_tree = dict()

        def evaluate(data, depth=0, current="IN", prior=(0, "IN")):
            if all(
                (
                    data.shape[0] > 1,
                    data[outcome].unique().shape[0] > 1,
                    depth <= max_depth if max_depth is not None else depth <= 1e6,
                )
            ):
                self.rule, sample_stats, self.score = self._run_solve(
                    data, outcome, target
                )
                self.size = sample_stats["size"]
                self.mean = sample_stats["mean"]

                if self.score > min_score:
                    key = prior + (depth, current)
                    self.rule_tree[key] = {
                        "rule": self.rule,
                        "score": self.score,
                        "size": self.size,
                        "mean": self.mean,
                    }

                    rule_data = self.get_rule_data(data).copy()
                    xrule_data = self.get_rule_data(data, xrule=True).copy()

                    try:
                        evaluate(rule_data, depth + 1, "IN", key)
                    except NotEnoughPointsError:
                        pass
                    try:
                        if (key + (depth + 1, "IN")) in self.rule_tree:
                            evaluate(xrule_data, depth + 1, "OUT", key)
                    except NotEnoughPointsError:
                        pass

        evaluate(mdf)

    def fit_meta(self, m_X, m_y, exp: BaseExploration):
        """
        Discover feature rules with the provided data.

        Parameters
        ----------
        m_X: MondoDataFrame of shape (n_samples, n_features)
            Where n_samples is the number of samples and n_features is the number
            of features

        m_y: MondoSeries of shape (n_samples,)
            Where n_samples is the number of samples and aligns with n_samples of m_X

        exp: an object that extends core.meta.BaseExploration
            The type of exploration used to discover feature rules

        """

        explorable_vars = [col for col in m_X.columns if col != m_y.name]
        m_X = m_X[explorable_vars].copy()

        mdf = MondoDataFrame(pd.concat((MondoDataFrame(m_y), m_X), axis=1))
        outcome = m_y.name

        encoder = DDTransformer(mdf)
        edf = utilities.encode_dataframe(mdf, encoder)
        outcome_encoded = encoder.original_to_encoded_column(outcome)
        exp.encode_options(mdf, outcome, encoder)

        edf, _ = utilities.sample_if_needed(edf, None, outcome_encoded)

        data = json.loads(edf.to_json(orient="records"))
        timeout = self._get_timeout(
            edf.shape[0], edf.shape[1] - 1, max_cycles=self.max_cycles
        )
        options = {
            "outcome": outcome_encoded,
            "data": data,
            "min_size_frac": self.min_size_frac,
            "timeout": timeout,
        }

        exp.explore(options)

        exp.decode_results(edf, outcome_encoded, encoder)

        rules = exp.result["rules"]
        meta_mdf = meta_utils.apply_meta_rules_to_df(rules, mdf)

        self.meta = {"rules": rules, "df": meta_mdf}

    def cross_validate(self, m_X, m_y, cv=3):
        """
        Cross validate rules against the provided data using a leave-one-out
            cross validation technique.

        Parameters
        ----------

        m_X: MondoDataFrame of shape (n_samples, n_features)
            Where n_samples is the number of samples and n_features is the
            number of features.

        m_y: MondoSeries of shape (n_samples,)
            Where n_samples is the number of samples and aligns with
            n_samples of m_X.

        cv: int or sklearn.model_selection object, optional (default=3)
            The number of folds (n_splits) used in cross validation, or a Sci-Kit
            Learn model splitter, such as StratifiedKFold, ShuffleSplit, etc.  If a
            number is given, the KFold technique is used.

        Returns
        -------

        scores: dict of float arrays of shape (n_splits,).  The possible keys for
            this dictionary are as follows:
            - train_scores: an array of scores when finding rules on validation sets
            - train_sizes: an array of sizes of rules on the training sets
            - test_scores: an array of scores when applying the rule on the test set
            - test_sizes: an array of sizes of rules on the test sets
            - rules: an array of rules (see solver.rule for object definition)
            - avg_loss: the mean squared error between train and test scores

        """

        if isinstance(cv, int) or isinstance(cv, float):
            cv = KFold(n_splits=cv, shuffle=True)

        X_train, X_test, y_train, y_test = train_test_split(
            m_X, m_y, test_size=0.33, random_state=1337
        )
        y_train.target_class = y_test.target_class = m_y.target_class

        mdf = MondoDataFrame(pd.concat((MondoDataFrame(y_test), X_test), axis=1))

        # For use in scoring later (_run_solve handles in-flight encoding)
        encoder = DDTransformer(mdf)
        edf = utilities.encode_dataframe(mdf, encoder)

        train_scores, train_sizes = [], []
        test_scores, test_sizes = [], []
        rules = []

        for train_index, test_index in cv.split(X_train, y_train):
            X_train_curr = X_train.iloc[train_index]
            y_train_curr = y_train.iloc[train_index]

            mdf_cur = MondoDataFrame(
                pd.concat((MondoDataFrame(y_train_curr), X_train_curr), axis=1)
            )

            outcome = m_y.name
            target = m_y.target_class

            train_rule, train_sample_stats, train_score = self._run_solve(
                mdf_cur, outcome, target
            )
            train_size = train_sample_stats["size"]

            if m_y.var_type == "discrete":
                # Discrete Stats - operate in original space
                mdf_applied = utilities.apply_rule_to_df(mdf, train_rule)

                test_sample_stats = utilities.get_stats(mdf_applied, outcome, target)
                test_score = utilities.score_rule(mdf, train_rule, outcome, target)
                test_size = test_sample_stats["size"]
            else:
                # Continous Stats - operate in encoded space
                test_rule = utilities.encode_rule(train_rule, encoder)
                outcome_encoded = encoder._column_encoder[outcome]

                edf_applied = utilities.apply_rule_to_df(edf, test_rule)

                test_sample_stats = utilities.get_stats(
                    edf_applied, outcome_encoded, target
                )
                test_score = utilities.score_rule(
                    edf, test_rule, outcome_encoded, target
                )
                test_size = test_sample_stats["size"]

            train_scores.append(train_score)
            train_sizes.append(train_size)
            test_scores.append(test_score)
            test_sizes.append(test_size)
            rules.append(train_rule)

        mse = np.mean((np.array(train_scores) - np.array(test_scores)) ** 2)

        result = {
            "train_scores": train_scores,
            "train_sizes": train_sizes,
            "test_scores": test_scores,
            "test_sizes": test_sizes,
            "rules": rules,
            "avg_loss": mse,
        }

        return result



from io import BytesIO
from typing import Dict, List, Union

from mondobrain.api.requestor import APIRequestor


def solve_start(
    outcome: str = None,
    target: str = None,
    data: List[Dict[str, Union[None, str, int, float]]] = None,
    timeout: int = None,
    min_size_frac: float = None,
    case_points: dict = None,
    **kwargs
):
    """Used for starting a solve using inline data (array of dicts)

    Parameters
    ----------
    outcome: str
        The feature to use as your outcome

    target: str
        The target for the outcome feature

    data: List[dict]
        The rows of data to use

    timeout: int (optional)
        The timeout to apply to the Solver loop. Defaults to API value

    min_size_frac: float (optional)
        Value between 0.0 and 1.0 that defines the minimum number of points needed for
        a valid rule discovered by the API

    case_points: dict
        Dictionary definining case points to include in rule
    """
    params = {
        "outcome": outcome,
        "target": target,
        "data": data,
        "timeout": timeout,
        "min_size_frac": min_size_frac,
        "case_points": case_points,
    }

    requestor = APIRequestor(**kwargs)

    return requestor.request("post", "solve.start", params=params)


def solve_start_file(
    outcome: str = None,
    target: str = None,
    data: BytesIO = None,
    timeout: int = None,
    min_size_frac: float = None,
    case_points: dict = None,
    **kwargs
) -> dict:
    """Used for starting a solve using a file to "carry" the data

    Parameters
    ----------
    outcome: str
        The feature to use as your outcome

    target: str
        The target for the outcome feature

    data: io.BytesIO
        A parquet buffer of the dataset
    """
    params = {
        "outcome": outcome,
        "target": target,
        "data": data,
        "timeout": timeout,
        "min_size_frac": min_size_frac,
        "case_points": case_points,
    }

    requestor = APIRequestor(**kwargs)
    supplied_headers = {"Content-Type": "multipart/form-data"}
    return requestor.request(
        "post", "solve.start.file", params=params, headers=supplied_headers
    )


def solve_result(id: str, **kwargs):
    """Read solve results (rules) from the API

    Parameters
    ----------
    id: str
        The task ID as returned from **api.solve_start** or **api.solve_start_file**
    """
    params = {"id": id}

    requestor = APIRequestor(**kwargs)

    return requestor.request("post", "solve.result", params=params)



import io
from typing import List

import numpy as np
import pandas as pd

from mondobrain.api import solve_result, solve_start_file
import mondobrain.datautils as datautils
from mondobrain.dd_transformer import DDTransformer
from mondobrain.error import InvalidTargetError, NotEnoughPointsError
from mondobrain.metrics import score
from mondobrain.utils import constants, utilities
from mondobrain.utils.data import is_discrete


def solve(
    df: pd.DataFrame,
    outcome: str = None,
    target: str = None,
    encode=True,
    sample=True,
    random_state=None,
    **kwargs,
):
    """Run a solve without worrying about API requests.

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to use for the solve

    outcome : str, default=None
        Which column to use as the outcome for the solve (sometimes referred to as a
        target feature). If None, then the first column is selected

    target: str, default=None
        The class in the outcome column to target. For a continuous outcome this should
        either be `min` or `max`. If None, then the first class or `max` is selected.

    encode: bool, default=True
        Whether or not the data should be encoded before being sent to the MB API.
        Encoding can result in additional time on client side. Disable if your data is
        largely non-sensitive.

    sample: bool, default=True
        Whether or not the data should be sampled before being sent to the MB API.
        Not pre-sampling the data can cause size limits to be reached and excessive
        solve times

    random_state : int or np.random.RandomStateInstance, default: 0
        Pseudo-random number generator to control the sampling state.
        Use an int for reproducible results across function calls.
        See the sklearn for more details.

    kwargs: any
        Remaining kwargs are sent so solve_start_file

    Returns
    -------
    rule: dict
        The conditions that the MB API found
    """
    if outcome is None:
        outcome = df.columns[0]

    if target is None:
        if df[outcome].dtype == np.object:
            target = df[outcome].iloc[0]
        else:
            target = "max"

    if target is np.nan:
        raise InvalidTargetError("Target cannot be NaN value")

    if encode:
        # Some utilities require original values
        df_orig = df.copy()
        outcome_orig = outcome
        target_orig = target

        encoder = DDTransformer(df)
        df = utilities.encode_dataframe(df, encoder)
        outcome = encoder.original_to_encoded_column(outcome)
        target = utilities.encode_value(df_orig, outcome_orig, target_orig, encoder)

    if sample:
        df = datautils.sample(
            df,
            2500,
            outcome=outcome,
            target=target,
            floor=3500,
            random_state=random_state,
        )

    _dimensionality_check(df, outcome, target)

    data = io.BytesIO()
    df.to_parquet(data)
    data.seek(0)  # Reset the pointer as `to_parquet` leaves it at the end

    task = solve_start_file(outcome=outcome, target=target, data=data, **kwargs)
    result = solve_result(id=task["id"])

    rule = result["rule"]

    if encode:
        # We need to decode the rule if we encoded
        rule = utilities.decode_rule(rule, encoder)

    return rule


def _dimensionality_check(df: pd.DataFrame, outcome: str, target: str):
    col = df[outcome]

    if is_discrete(col):
        size = col[col == target].shape[0]
    else:
        if col.std() == 0:
            raise NotEnoughPointsError("Outcome column has no variance")

        size = col.shape[0]

    if size <= constants.MIN_SOLVER_SIZE:
        raise NotEnoughPointsError("Not enough points")


def _exhaustive_single(df: pd.DataFrame, outcome: str, target: str, **kwargs) -> List:
    try:
        rule = solve(df, outcome=outcome, target=target, **kwargs)
    except NotEnoughPointsError:
        return []

    try:
        df_inverse = datautils.apply_rule(df, rule, inverse=True)
    except ValueError:
        df_inverse = df.loc[[]]

    if df_inverse.size > 0:
        return [rule] + _exhaustive_single(df_inverse, outcome, target, **kwargs)
    else:
        return []


def _exhaustive_best(df: pd.DataFrame, outcome: str, **kwargs) -> List:
    best_score = -99999.0
    for target in df[outcome].unique():
        try:
            rule = solve(df, outcome=outcome, target=target, **kwargs)
            if rule:
                rule_score = score(df, outcome, target, rule)
            else:
                rule_score = best_score
            if rule_score > best_score:
                best_rule = rule
                best_target = target
            else:
                continue
        except NotEnoughPointsError:
            return []

        try:
            df_inverse = datautils.apply_rule(df, best_rule, inverse=True)
        except ValueError:
            df_inverse = df.loc[[]]

        if df_inverse.size > 0:
            return [[best_rule, best_target]] + _exhaustive_best(
                df_inverse, outcome, **kwargs
            )
        else:
            return []


def _exhaustive_any(df: pd.DataFrame, outcome: str, **kwargs) -> List:
    rule_set = []
    targets = df[outcome].unique()
    for target in targets:
        try:
            rule = solve(df, outcome=outcome, target=target, **kwargs)
            rule_set.append(rule)
        except NotEnoughPointsError:
            return []

    try:
        df_inverse = datautils.apply_rules(df, rule_set, inverse=True)
    except ValueError:
        df_inverse = df.loc[[]]

    if df_inverse.size > 0:
        return [rule_set] + _exhaustive_any(df_inverse, outcome, **kwargs)
    else:
        return []


def exhaustive_solve(
    df: pd.DataFrame,
    outcome: str = None,
    target: str = None,
    approach: str = "single",
    **solve_params,
):
    """
        Build a set of conditions that exhaustively cover the entirety of a dataset

        Parameters
        ----------
        df: Pandas DataFrame
            Data values for the exhaustive solve containing Outcome Feature
            and Independent Feature variables.

        outcome : str
            The label of the dependent variable used for learning.

        target : str (optional), default=None
            The target class of the outcome. Applicable when running a
            exhaustive solve using the 'single' approach.

        approach : str (optional), default='single'
            The approach to use for exhaustive solves. Approaches have a different
            return shape.

            - ‘single’ (default) : (n_solves,)
                Search for rules on an individual target in a dataset.
            - ‘best’ : (2, n_solves)
                Generates rules for all targets, chooses the best rule and removes
                points in that rule. Repeats for remaining data.
            - ‘any’ : (n_classes, n_solves)
                Generates rules for all targets and removes points in any rule. Repeats
                for remaining data.

        **solve_params : any
            Parameters to pass on to `solve`

        Returns
        -------
        rule_set: List
            The set of conditions that the MB API found covering all points
            in the provided dataset
    """
    if outcome is None:
        outcome = df.columns[0]

    if approach == "single":
        return _exhaustive_single(df, outcome, target, **solve_params)
    if approach == "best":
        return _exhaustive_best(df, outcome, **solve_params)
    if approach == "any":
        return _exhaustive_any(df, outcome, **solve_params)

    raise ValueError(f"approach {approach} does not exist")



from auth0.v3.authentication import GetToken
from requests.packages.urllib3.fields import RequestField
from requests.packages.urllib3.filepost import encode_multipart_formdata

import mondobrain


def get_token(key, secret, auth0_domain=None) -> str:
    domain = auth0_domain or mondobrain.auth0_domain

    gt = GetToken(domain)
    resp = gt.client_credentials(key, secret, "https://api.mondobrain.com/")
    access = resp.get("access_token", None)

    return access


def encode_multipart(params: dict):
    new_fields = []

    for key, value in params.items():
        if value is None:
            continue

        if hasattr(value, "read"):
            filename = "data"
            if hasattr(value, "name"):
                filename = str(value.name)

            fdata = value.read()
            rf = RequestField(name=key, data=fdata, filename=filename)
            rf.make_multipart(content_type="application/octet-stream")
            new_fields.append(rf)
        else:
            if not isinstance(value, bytes):
                value = str(value)

            key = key.decode("utf-8") if isinstance(key, bytes) else key
            value = value.encode("utf-8") if isinstance(value, str) else value

            new_fields.append((key, value))

    return encode_multipart_formdata(new_fields)



from enum import Enum

import mondobrain as mb
from mondobrain.core.frame import MondoDataFrame
from mondobrain.dd_transformer import DDTransformer
from mondobrain.utils import meta as meta_utils, utilities


class Direction(str, Enum):
    in_ = "in"
    out = "out"
    both = "both"


class Inverse(str, Enum):
    in_rule = "in_rule"


class BaseExploration:
    def __init__(self):
        self.result = None

    def explore(self, options: dict):
        raise NotImplementedError(".explore() must be overridden")

    def encode_options(self, mdf: MondoDataFrame, outcome: str, encoder: DDTransformer):
        pass

    def decode_results(self, edf: MondoDataFrame, outcome: str, encoder: DDTransformer):
        if self.result is not None:
            self.result["rules"] = meta_utils.decode_meta_rules(
                self.result["rules"], edf, outcome, encoder
            )


class ClassExploration(BaseExploration):
    """
    Define a MetaMondo exploration using a class-based strategy.

    Feature rules are gathered by learning on multiple classes in multiple directions.
    For discrete outcome variables, this means in or out of each outcome modality and
    for continuous outcome variables, this means min and/or max

    **Parameters**

    classes: List[str] (optional), default=None
        For discrete outcomes, a list of modalities to explore.
        For continuous outcomes, a list containing 'min' and/or 'max'.
        If None, explore all classes.

    direction: 'in', 'out', 'both' (optional), default='both'
        Direction to explore each outcome class.  Only applicable for discrete
        outcomes.

        * **in**: find feature rules inside of the outcome class
        * **out**: find feature rules outside of the outcome class
        * **both**: do both in and out
        * **None**: do not explore directions (for continuous outcome variables)

    inverse: 'in_rule' (optional), default='in_rule'
        Strategy for finding "inverse" rules when a feature rule is found

        * **in_rule**: find a rule inside the feature rule in the opposite direction
        * **None**: do not add inverse rules

    depth: int (optional), default=1
        The number of exploration cycles.

    parallel: bool (optional), default=False
        Explore each outcome class independently and in parallel.

    **Examples**

    Using the solver to run a ClassExploration

    >>> mb.api_key = "<API KEY>"
    >>> mb.api_secret = "<API SECRET>"
    >>> solver = mb.Solver()
    >>> solver.fit_meta(
    >>>     features, outcome,
    >>>     exp=ClassExploration(classes=['outcomeA'], inverse=None, depth=3)
    >>> )
    >>> solver.meta
    {
        'rules': ...list of rules found by MetaMondo...,
        'df': ...MondoDataFrame with feature rules included...
    }
    """

    def __init__(
        self,
        classes=None,
        direction=Direction.both,
        inverse=Inverse.in_rule,
        depth=1,
        parallel=False,
    ):
        super(ClassExploration, self).__init__()
        self.classes = classes
        self.direction = direction
        self.inverse = inverse
        self.depth = depth
        self.parallel = parallel

    def explore(self, options: dict):
        options.update(
            {
                "classes": self.classes,
                "direction": self.direction,
                "inverse": self.inverse,
                "depth": self.depth,
                "parallel": self.parallel,
            }
        )

        task = mb.api.meta_class_start(**options)

        self.result = mb.api.meta_result(id=task["id"])

    def encode_options(self, mdf: MondoDataFrame, outcome: str, encoder: DDTransformer):
        if mdf[outcome].var_type == "discrete" and self.classes is not None:
            self.classes = [
                utilities.encode_value(mdf, outcome, x, encoder) for x in self.classes
            ]


class ExclusionSetExploration(BaseExploration):
    """
    Define a MetaMondo exploration using an exclusion set strategy.

    Feature rules are gathered by learning a rule, then learning again on the exclusion
    set (the set of data points outside of the rule), repeating until the exclusion set
    reaches a minimum number of data points.

    **Parameters**

    min_size: int (optional), default=10
        The minimum size of the exclusion set.  Stops exploration when the
        exclusion set is below this size.

    classes: List[str] (optional), default=None
        For discrete outcomes, a list of modalities to explore.
        For continuous outcomes, a list containing 'min' and/or 'max'.
        If None, explore all classes.

    direction: 'in', 'out', 'both' (optional), default='both'
        Direction to explore each outcome class.  Only applicable for discrete
        outcomes.

        * **in**: find feature rules inside of the outcome class
        * **out**: find feature rules outside of the outcome class
        * **both**: do both in and out
        * **None**: do not explore directions (for continuous outcome variables)

    depth: int (optional), default=1
        The number of exploration cycles.

    parallel: bool (optional), default=False
        Explore each outcome class independently and in parallel.

    **Examples**

    Using the solver to run an ExclusionSetExploration

    >>> mb.api_key = "<API KEY>"
    >>> mb.api_secret = "<API SECRET>"
    >>> solver = mb.Solver()
    >>> solver.fit_meta(
    >>>     features, outcome,
    >>>     exp=ExclusionSetExploration(min_size=5, direction='in', parallel=True)
    >>> )
    >>> solver.meta
    {
        'rules': ...list of rules found by MetaMondo...,
        'df': ...MondoDataFrame with feature rules included...
    }
    """

    def __init__(
        self,
        min_size=10,
        classes=None,
        direction=Direction.both,
        depth=1,
        parallel=False,
    ):
        super(ExclusionSetExploration, self).__init__()
        self.min_size = min_size
        self.classes = classes
        self.direction = direction
        self.depth = depth
        self.parallel = parallel

    def explore(self, options: dict):
        options.update(
            {
                "min_size": self.min_size,
                "classes": self.classes,
                "direction": self.direction,
                "depth": self.depth,
                "parallel": self.parallel,
            }
        )

        task = mb.api.meta_exclusion_set_start(**options)

        self.result = mb.api.meta_result(id=task["id"])

    def encode_options(self, mdf: MondoDataFrame, outcome: str, encoder: DDTransformer):
        if mdf[outcome].var_type == "discrete" and self.classes is not None:
            self.classes = [
                utilities.encode_value(mdf, outcome, x, encoder) for x in self.classes
            ]



from numpy import issubdtype, number
from pandas import Series

from mondobrain.error import InvalidTargetError


class MondoSeries(Series):
    """
    A MondoSeries object is a pandas.Series that has certain
    MondoBrain specific operations and methods.

    **Parameters**

    data: array-like, Iterable, dict, or scalar value
        Contains data stored in Series.

    **Attributes**

    target_class: str (optional for feature variables)
        If the MondoSeries is passed to the prescriber.Solver object as
        a target variable, a target_class must be defined for the
        MondoBrain Solver to optimize with.

        Discrete variables: label set from the set prescriber.Solver.classes
        Continous variables: "max" or "min"

    classes: numpy.array (read only)
        A list of class labels or modalities known to the MondoSeries.

    case_point: numeric (optional)
        Scalar value of type int or float that can be set for continuous
        variables. Case point ensures that a rule is found with the Solver
        that contains the referenced point.

    var_type: str (read only)
        Label that identifies the variable as "discrete" for categorical values
        and "continuous" for numeric values.

    **Examples**

    >>> from mondobrain.MondoSeries import MondoSeries
    >>> import Numpy as np
    >>> arr = np.arange(100)
    >>> ms = MondoSeries(arr, name='some_variable')
    >>> ms.target_class = 'some_class'

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.case_point = None
        self.target_class = None

    @property
    def _constructor(self):
        return MondoSeries

    @property
    def target_class(self):
        return self._target_class

    @target_class.setter
    def target_class(self, value):
        if value is None:
            self._target_class = None

        elif self.var_type == "discrete":
            if value in self.classes:
                self._target_class = value
            else:
                raise InvalidTargetError(
                    f'"{value}" is not a modality of the variable "{self.name}"'
                )

        else:
            if value in ["max", "min"]:
                self._target_class = value
            else:
                raise InvalidTargetError(
                    f'Outcome target must be either "max" or "min" '
                    f'for the continuous variable "{self.name}"'
                )

    @property
    def classes(self):
        return self.unique()

    @classes.setter
    def classes(self, value):
        raise TypeError("'target_class' attribute does not support assignment")

    @property
    def case_point(self):
        return self._case_point

    @case_point.setter
    def case_point(self, value):
        self._case_point = value

    @property
    def var_type(self):
        _var_type = None
        if issubdtype(self.dtype, number):
            _var_type = "continuous"
        else:
            _var_type = "discrete"
        return _var_type



import json
import platform

import mondobrain
from mondobrain import error, utils
from mondobrain.api import clients
from mondobrain.api.util import encode_multipart, get_token


class APIRequestor(object):
    def __init__(self, token=None, client=None, api_base=None):
        self.api_base = api_base or mondobrain.api_base
        self.api_token = token or mondobrain.api_token

        if token:
            self.api_token = token
        elif mondobrain.api_token:
            self.api_token = mondobrain.api_token
        else:
            mondobrain.api_token = get_token(mondobrain.api_key, mondobrain.api_secret)
            self.api_token = mondobrain.api_token

        if client:
            self._client = client
        elif mondobrain.default_http_client:
            self._client = mondobrain.default_http_client
        else:
            mondobrain.default_http_client = clients.new_default_http_client()
            self._client = mondobrain.default_http_client

    def request(self, method, url, params=None, headers=None):
        rbody, rcode, rheaders = self.request_raw(method.lower(), url, params, headers)

        resp = self.interpret_response(rbody, rcode, rheaders)
        return resp

    def handle_error_response(self, rbody, rcode, resp, rheaders):
        try:
            error_data = {"type": resp["type"], "message": resp["message"]}

            code = resp.get("code", None)
            if code is not None:
                error_data["code"] = code
        except (KeyError, TypeError):
            raise error.APIError(
                "Invalid response object from API: %r "
                "(HTTP response code was %d)" % (rbody, rcode),
                rbody,
                rcode,
                resp,
                rheaders,
            )

        msg = error_data.get("message", None)

        if rcode == 401:
            raise error.AuthenticationError(msg, rbody, rcode, resp, rheaders, code)

        elif rcode == 402:
            eType = error_data["type"]
            raise error.ProcessingError(msg, rbody, rcode, resp, rheaders, code, eType)

        elif rcode == 403:
            raise error.PermissionError(msg, rbody, rcode, resp, rheaders, code)

        else:
            raise error.APIError(msg, rbody, resp, rheaders)

    def request_headers(self, api_token):
        user_agent = "Mondobrain/v1 PythonBindings/%s" % (mondobrain.__version__,)

        ua = {
            "bindings_version": mondobrain.__version__,
            "lang": "python",
            "publisher": "mondobrain",
            "httplib": "requests",
            "lang_version": platform.python_version(),
            "platform": platform.platform(),
            "uname": " ".join(platform.uname()),
        }

        headers = {
            "X-Mondobrain-Client-User-Agent": json.dumps(ua),
            "User-Agent": user_agent,
            "Authorization": "Bearer %s" % (api_token,),
            "Content-Type": "application/json",
        }

        return headers

    def request_raw(self, method, url, params=None, supplied_headers=None):
        """
        Mechanism for issuing an API call
        """

        # TODO: token flow
        my_api_token = self.api_token

        abs_url = "%s%s" % (self.api_base, url)

        if (
            supplied_headers is not None
            and supplied_headers.get("Content-Type") == "multipart/form-data"
        ):
            post_data, content_type = encode_multipart(params)
            supplied_headers["Content-Type"] = content_type
        else:
            params = {k: v for k, v in params.items() if v is not None}
            post_data = json.dumps(params)

        headers = self.request_headers(my_api_token)
        if supplied_headers is not None:
            for key, value in supplied_headers.items():
                headers[key] = value

        utils.log_info("Request to Mondobrain api", method=method, path=abs_url)
        utils.log_debug("Post details", post_data=params)

        rbody, rcode, rheaders = self._client.request_with_polling(
            method, abs_url, headers, post_data
        )

        utils.log_info("Mondobrain API response", path=abs_url, response_code=rcode)
        utils.log_debug("API response body", body=rbody)

        return rbody, rcode, rheaders

    def interpret_response(self, rbody, rcode, rheaders):
        try:
            if hasattr(rbody, "decode"):
                rbody = rbody.decode("utf-8")

            data = json.loads(rbody)
        except Exception:
            raise error.APIError(
                "Invalid response body from API: %s "
                "(HTTP response code was %d)" % (rbody, rcode),
                rbody,
                rcode,
                headers=rheaders,
            )

        if not 200 <= rcode < 300:
            self.handle_error_response(rbody, rcode, data, rheaders)

        return data



from IPython.display import Image
import networkx as nx
import pygraphviz as pgv


def create_rule_graph(rules):
    G = nx.DiGraph()

    for node in rules:
        G.add_node(node, rule=rules[node]["rule"])

    for node in G.nodes:
        if node[:-2] in G.nodes:
            lbl = "score: {:f}\nsize: {}".format(
                rules[node]["score"], rules[node]["size"]
            )
            G.add_edge(node[:-2], node, label=lbl)

    lbl = "score: {:f}\nsize: {}".format(
        rules[(0, "IN", 0, "IN")]["score"], rules[(0, "IN", 0, "IN")]["size"]
    )
    G.add_edge((0, "IN", 0, "IN"), (0, "IN", 0, "IN"), label=lbl)

    return G


def visualize_rule_graph(G):
    nx.nx_agraph.write_dot(G, "rules.dot")
    G = pgv.AGraph("rules.dot", directed=True)

    G.node_attr["shape"] = "box"
    G.edge_attr["color"] = "red"

    G.layout("dot")
    G.draw("graph.png")

    return Image(filename="graph.png")


def visualize_rule_tree(rule_tree):
    return visualize_rule_graph(create_rule_graph(rule_tree))



from .requestor import APIRequestor


def api_test(error: str = None, **kwargs):
    """Tests the ability to connect to the API
    """
    params = {}
    if error:
        params["error"] = error

    requestor = APIRequestor(**kwargs)

    return requestor.request("post", "api.test", params=params)



import numpy as np
import pandas as pd


def target_metrics(s: pd.Series, target: str = None) -> dict:
    """ Get metrics for a target within a column

    Parameters
    ----------
    s : pd.Series
        The series to calculate metrics for

    target: str, default=None
        The class in s to calculate metrics for. For a continuous outcome this should
        either be `min` or `max`. If None, then the first class or `max` is selected.
    """
    if target is None:
        target = s[0] if s.dtype == np.object else "max"

    if s.dtype == np.object:
        values = (s == target).astype(np.int)
    else:
        values = s
        if target == "min":
            values *= -1

    mean = values.mean()
    above_mean = int((values >= mean).sum()) if mean > 0 else 0

    return {
        "mean": mean,
        "std": values.std(ddof=0),
        "size": int(values.size),
        "size_above_mean": above_mean,
    }



import json
import random
import textwrap
import threading
import time

import requests

import mondobrain
from mondobrain import error, utils


def _now_ms():
    return int(round(time.time() * 1000))


def new_default_http_client(*args, **kwargs):
    return RequestsClient(*args, **kwargs)


class HTTPClient(object):
    MAX_DELAY = 2
    INITIAL_DELAY = 0.5
    MAX_RETRY_AFTER = 60

    def __init__(self):
        self._thread_local = threading.local()

    def request_with_polling(self, method, url, headers, post_data=None):
        """Handles retrying requests that are known to require polling
        """
        num_polls = 0

        while True:
            try:
                response = self.request_with_retries(method, url, headers, post_data)
                connection_error = None
            except error.APIConnectionError as e:
                connection_error = e
                response = None

            if self._should_poll(response):
                num_polls += 1
                sleep_time = self._sleep_time_seconds(num_polls, response)
                utils.log_info(
                    (
                        "Initiating poll %i for request %s %s after "
                        "sleeping %.2f seconds." % (num_polls, method, url, sleep_time)
                    )
                )

                time.sleep(sleep_time)

            else:
                if response is not None:
                    return response
                else:
                    raise connection_error

    def request_with_retries(self, method, url, headers, post_data=None):
        """
            Handles network retries.
        """
        self._add_telemetry_header(headers)

        num_retries = 0

        while True:
            try:
                response = self.request(method, url, headers, post_data)
                connection_error = None
            except error.APIConnectionError as e:
                connection_error = e
                response = None

            if self._should_retry(response, connection_error, num_retries):
                if connection_error:
                    utils.log_info(
                        "Encountered a retryable error %s" % str(connection_error)
                    )

                num_retries += 1
                sleep_time = self._sleep_time_seconds(num_retries, response)
                utils.log_info(
                    (
                        "Initiating retry %i for request %s %s after "
                        "sleeping %.2f seconds."
                        % (num_retries, method, url, sleep_time)
                    )
                )

                time.sleep(sleep_time)
            else:
                if response is not None:
                    return response
                else:
                    raise connection_error

    def request(self, method, url, headers, post_data=None):
        raise NotImplementedError("HTTPClient subclasses must implement `request`")

    def _should_retry(self, response, api_connection_error, num_retries):
        if num_retries >= self._max_network_retries():
            return False

        if response is None:
            return api_connection_error.should_retry

        _, status_code, rheaders = response

        if rheaders is not None and "mondobrain-should-retry" in rheaders:
            if rheaders["mondobrain-should-retry"] == "false":
                return False
            if rheaders["mondobrain-should-retry"] == "true":
                return True

        if status_code >= 500:
            return True

        return False

    def _should_poll(self, response):
        if not response:
            return False

        _, status_code, _ = response

        if status_code == 202:
            return True

        return False

    def _max_network_retries(self):
        from mondobrain import max_network_retries

        return max_network_retries

    def _retry_after_header(self, response=None):
        if response is None:
            return None

        _, _, rheaders = response

        try:
            return int(rheaders["retry-after"])
        except (KeyError, ValueError):
            return None

    def _sleep_time_seconds(self, num_retries, response=None):
        sleep_seconds = min(
            HTTPClient.INITIAL_DELAY * (2 ** (num_retries - 1)), HTTPClient.MAX_DELAY
        )

        sleep_seconds = self._add_jitter_time(sleep_seconds)

        # But never sleep less than initial delay
        sleep_seconds = max(HTTPClient.INITIAL_DELAY, sleep_seconds)

        # And never sleep less than the time the API asks us to wait,
        # assuming it's a reasonable ask.
        retry_after = self._retry_after_header(response) or 0
        if retry_after <= HTTPClient.MAX_RETRY_AFTER:
            sleep_seconds = max(retry_after, sleep_seconds)

        return sleep_seconds

    def _add_jitter_time(self, sleep_seconds):
        # Randomize the value in [(sleep_seconds/ 2) to (sleep_seconds)]
        # Also separated method here to isolate randomness for tests
        sleep_seconds *= 0.5 * (1 + random.uniform(0, 1))
        return sleep_seconds

    def _add_telemetry_header(self, headers):
        last_request_metrics = getattr(self._thread_local, "last_request_metrics", None)
        if mondobrain.enable_telemetry and last_request_metrics:
            telemetry = {"last_request_metrics": last_request_metrics}
            headers["X-Mondobrain-Client-Telemetry"] = json.dumps(telemetry)

    def _record_request_metrics(self, response, request_start):
        _, _, rheaders = response
        if "X-Request-Id" in rheaders and mondobrain.enable_telemetry:
            request_id = rheaders["X-Request-Id"]
            request_duration_ms = _now_ms() - request_start
            self._thread_local.last_request_metrics = {
                "request_id": request_id,
                "request_duration_ms": request_duration_ms,
            }

    def close(self):
        raise NotImplementedError("HTTPClient subclasses must implement `close`")


class RequestsClient(HTTPClient):
    name = "requests"

    def __init__(self, timeout=80, session=None, **kwargs):
        super(RequestsClient, self).__init__(**kwargs)
        self._session = session
        self._timeout = timeout

    def request(self, method, url, headers, post_data=None):
        kwargs = {}
        if getattr(self._thread_local, "session", None) is None:
            self._thread_local.session = self._session or requests.Session()

        try:
            try:
                result = self._thread_local.session.request(
                    method,
                    url,
                    headers=headers,
                    data=post_data,
                    timeout=self._timeout,
                    **kwargs
                )
            except TypeError as e:
                raise TypeError(
                    "Warning: It looks like your installed version of the "
                    '"requests" library is not compatible with Mondobrain\'s '
                    "usage thereof. (HINT: The most likely cause is that "
                    'your "requests" library is out of date. You can fix '
                    'that by running "pip install -U requests".) The '
                    "underlying error was: %s" % (e,)
                )

            content = result.content
            status_code = result.status_code
        except Exception as e:
            self._handle_request_error(e)

        return content, status_code, result.headers

    def _handle_request_error(self, e):
        # Catch SSL error first as it belongs to ConnectionError,
        # but we don't want to retry
        if isinstance(e, requests.exceptions.SSLError):
            msg = (
                "Could not verify Mondobrain's SSL certificate. Please make "
                "sure that your network is not intercepting certificates. "
                "If this problem persists, reach out to your representative"
            )

            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = False

        # Retry only timeout and connect errors; similar to urllib3 Retry
        elif isinstance(
            e, (requests.exceptions.Timeout, requests.exceptions.ConnectionError),
        ):
            msg = (
                "Unexpected error communicating with Mondobrain. "
                "If this problem persists, reach out to your representative"
            )

            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = True

        # Catch remaining request exceptions
        elif isinstance(e, requests.exceptions.RequestException):
            msg = (
                "Unexpected error communicating with Mondobrain. "
                "If this problem persists, reach out to your representative"
            )

            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = False

        else:
            msg = (
                "Unexpected error communicating with Mondobrain. "
                "It looks like there's probably a configuration "
                "issue locally.  If this problem persists, reach out"
                "to your representative"
            )
            err = "A %s was raised" % (type(e).__name__,)
            if str(e):
                err += " with error message %s" % (str(e),)
            else:
                err += " with no error message"
            should_retry = False

        msg = textwrap.fill(msg) + "\n\n(Network error %s)" % (err,)
        raise error.APIConnectionError(msg, should_retry=should_retry)

    def close(self):
        if getattr(self._thread_local, "session", None) is not None:
            self._thread_local.session.close()



class MondobrainError(Exception):
    pass


class APIError(MondobrainError):
    def __init__(
        self,
        message=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
    ):
        super(APIError, self).__init__(message)

        if http_body and hasattr(http_body, "decode"):
            try:
                http_body = http_body.decode("utf-8")
            except BaseException:
                http_body = "<Could not decode body as utf-8>"

        self._message = message
        self.http_body = http_body
        self.json_body = json_body
        self.headers = headers or {}
        self.code = code
        self.request_id = self.headers.get("x-request-id", None)

    def __str__(self):
        msg = self._message or "<empty message>"
        if self.request_id is not None:
            return u"Request {0}: {1}".format(self.request_id, msg)

        return msg

    def __repr__(self):
        return "%s(message=%r, http_status=%r, request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.http_status,
            self.request_id,
        )


class APIConnectionError(APIError):
    def __init__(
        self,
        message,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
        should_retry=False,
    ):
        super(APIConnectionError, self).__init__(
            message, http_body, http_status, json_body, headers, code
        )
        self.should_retry = should_retry


class AuthenticationError(APIError):
    pass


class PermissionError(APIError):
    pass


class ProcessingError(APIError):
    def __init__(
        self,
        message,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
        type=None,
    ):
        super(ProcessingError, self).__init__(
            message, http_body, http_status, json_body, headers, code
        )

        self.type = type


class InvalidTargetError(MondobrainError):
    pass


class NotEnoughPointsError(MondobrainError):
    pass



from sklearn.preprocessing import LabelEncoder
from sklearn.utils import column_or_1d

from mondobrain.utils.data import is_continuous


def _encode_array(y):
    y = column_or_1d(y)

    if not is_continuous(y.dtype):
        raise ValueError("only able to encode numeric arrays")

    enc = LabelEncoder()
    return enc.fit_transform(y)


def encoded_mean(y):
    return _encode_array(y).mean()



import numpy as np
import pandas as pd

from mondobrain.utils.data import is_discrete


def _uniform_stratified_sample(
    df: pd.DataFrame, size: int, outcome: str, random_state=None
) -> pd.DataFrame:
    return (
        df.groupby(outcome, group_keys=False)
        .apply(
            lambda x: x.sample(
                int(np.rint(size * x.shape[0] / df.shape[0])), random_state=random_state
            )
        )
        .sample(frac=1, random_state=random_state)
    )


def _oversampled_sample(
    df: pd.DataFrame, size: int, outcome: str, target: str, random_state=None
) -> pd.DataFrame:
    target_prob = 1 - (df[outcome] == target).sum() / df.shape[0]
    alternate_prob = 1 - target_prob

    pick = np.where(df[outcome] == target, target_prob, alternate_prob)

    weights = pick / pick.sum()

    return df.sample(size, random_state=random_state, weights=weights)


def sample(
    df: pd.DataFrame,
    size: int,
    outcome=None,
    target=None,
    floor: int = None,
    random_state=None,
):
    """Sample a dataframe while taking the outcome & target into consideration

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to sample

    outcome : str, default=None
        Which column to use as the outcome consideration (sometimes referred to as a
        target feature). If None, then the first column is selected

    target: str, default=None
        The class in the outcome column to consider. For a continuous outcome this has
        no effect. If None, all targets are considered.

    floor: int, default=None
        The lower limit at which we should consider sampling. This lower limit is
        inclusive. If the number of rows is lte (<=) the floor, we return a copy of the
        dataframe

    random_state : int or np.random.RandomStateInstance, default: 0
        Pseudo-random number generator to control the sampling state.
        Use an int for reproducible results across function calls.
        See the sklearn for more details.


    Returns
    -------
    rule: dict
        The conditions that the MB API found
    """
    if floor is not None and df.shape[0] <= floor:
        return df.copy()

    if outcome is None:
        outcome = df.columns[0]

    # For discrete outcomes, we might use oversampled outcomes and the total number of
    # points are less than 0.15 of the overall
    if is_discrete(df[outcome]):

        # use uniform stratified sampling when target is not specified
        if target is None:
            return _uniform_stratified_sample(
                df, size, outcome, random_state=random_state
            )

        # if target is less than 15%, oversample the target
        if df[outcome].value_counts()[target] / df.shape[0] < 0.15:
            return _oversampled_sample(
                df, size, outcome, target, random_state=random_state
            )

        # if outcome only has 2 classes, check alternate for oversample conditions
        if df[outcome].value_counts().shape[0] == 2:
            modalities = df[outcome].value_counts().index
            alternate = modalities[~np.in1d(modalities, target)][0]

            # if the alternate target is less than 15%, oversample it
            if df[outcome].value_counts()[alternate] / df.shape[0] < 0.15:
                return _oversampled_sample(
                    df, size, outcome, alternate, random_state=random_state
                )

    # Continuous outcomes get sampled regularly, as well as any discrete that don't
    # meet the above conditions
    return df.sample(size, random_state=random_state)



from typing import Dict, List, Union

import numpy as np
import pandas as pd


def _condition_mask(s: pd.Series, cond: Union[dict, str]):
    if isinstance(cond, dict):
        lo, hi = cond["lo"], cond["hi"]
        return (s >= lo) & (s <= hi)

    return s == cond


def conditions_mask(df: pd.DataFrame, rule: dict):
    """ Get a mask based on a dictionary of conditions
    """
    masks = []
    for key, cond in rule.items():
        masks.append(_condition_mask(df[key], cond))

    return np.all(np.array(masks), axis=0)


def apply_rule(df: pd.DataFrame, rule: dict, inverse: bool = False) -> pd.DataFrame:
    """ Applies the rule to the population df to obtain a sample

    Parameters
    ----------
    df : pd.DataFrame
        The dataset to apply the rule to

    rule : dict
        A set of conditions to apply to the dataframe

    inverse : boolean, default=False
        If True, the inverse of the rule is applied

    Returns
    -------
    sample: pd.DataFrame
        The sample of the original dataframe based on rule conditions
    """
    if not rule:
        raise ValueError("An empty rule cannot be applied to a dataframe")

    mask = conditions_mask(df, rule)

    if inverse:
        mask = ~mask

    return df[mask]


def apply_rules(
    df: pd.DataFrame, rules: List[Dict], inverse: bool = False, operation: str = "or"
) -> pd.DataFrame:
    mask = [conditions_mask(df, rule) for rule in rules if rule]

    if len(mask) == 0:
        raise ValueError("No rule found to apply to the dataframe")

    if operation == "or":
        mask = np.any(mask, axis=0)
    elif operation == "and":
        mask = np.all(mask, axis=0)
    else:
        raise ValueError("operation type not supported")

    if inverse:
        mask = ~mask

    return df[mask]



from .requestor import APIRequestor


def language_entities(message: str, metadata: dict, **kwargs):
    """Use the API to extract entities in a sentence
    """
    params = {
        "message": message,
        "metadata": metadata,
    }

    requestor = APIRequestor(**kwargs)
    return requestor.request("post", "language.entities", params=params)



import numpy as np
import pandas as pd
from sklearn.model_selection import KFold, train_test_split

from mondobrain.datautils import apply_rule
from mondobrain.metrics import score
from mondobrain.solver import solve


def cross_validate(df: pd.DataFrame, cv=3, random_state=1337, **kwargs):
    if isinstance(cv, (int, float)):
        cv = KFold(n_splits=cv, shuffle=True)

    train, test = train_test_split(df, test_size=0.33, random_state=random_state)

    rules = []
    train_sizes, train_scores = [], []
    test_sizes, test_scores = [], []

    for train_idx, _ in cv.split(train):
        train_curr = train.iloc[train_idx]

        rule = solve(train_curr, random_state=random_state, **kwargs)
        if not rule:
            continue

        rules.append(rule)

        # Get train metrics
        train_applied = apply_rule(train_curr, rule)

        train_size, _ = train_applied.shape
        train_score = score(train_applied, population=train_curr, **kwargs)

        train_sizes.append(train_size)
        train_scores.append(train_score)

        # Get test metrics
        test_applied = apply_rule(test, rule)

        test_size, _ = test_applied.shape
        test_score = score(test_applied, population=test, **kwargs)

        test_sizes.append(test_size)
        test_scores.append(test_score)

    mse = np.mean((np.array(train_scores) - np.array(test_scores)) ** 2)

    return {
        "avg_loss": mse,
        "train_scores": train_scores,
        "train_sizes": train_sizes,
        "test_scores": test_scores,
        "test_sizes": test_sizes,
        "rules": rules,
    }



import pandas as pd
import pkg_resources

from mondobrain.core.frame import MondoDataFrame


def __load_dataset(name):
    with pkg_resources.resource_stream(__name__, f"data/{name}.csv") as stream:
        df = pd.read_csv(stream)
    return MondoDataFrame(df)


def load_bank_risk_date():
    return __load_dataset("bank-credit-risk-with-date")


def load_bronchial_risk():
    return __load_dataset("bronchial-risk")


def load_iris():
    """Return a dataframe about Irises (the flower).

    Contains the following fields:
        SepalLength          float
        SepalWidth           float
        PetalLength          float
        PetalWidth           float
        Name                 str
    """
    return __load_dataset("iris")


def load_jets():
    return __load_dataset("jets")


def load_mazda():
    return __load_dataset("mazda")


def load_political_behavior():
    return __load_dataset("political-behavior")


def load_shampoo():
    return __load_dataset("shampoo")


def load_titanic():
    return __load_dataset("titanic")



import distutils.cmd
import os
import re
import sys

from setuptools import find_packages, setup

ROOT = os.path.dirname(__file__)

SEMVER_REGEX = r"(?P<semver>(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?)"  # noqa: E501
VERSION_RE = re.compile(r'__version__ = "' + SEMVER_REGEX + r'"')


def get_version():
    init = open(os.path.join(ROOT, "mondobrain", "__init__.py")).read()
    return VERSION_RE.search(init).group("semver")


class VerifyVersionCommand(distutils.cmd.Command):
    """Custom command to verify that the git tag matches our version"""

    description = "verify that the git tag matches our version"

    user_options = [
        ("git-tag=", None, "Tag to verify"),
        ("git-branch=", None, "Branch to verify"),
    ]

    def initialize_options(self):
        """Set default values for options."""
        # Each user option must be listed here with their default value.
        self.git_tag = os.getenv("CIRCLE_TAG")
        self.git_branch = os.getenv("CIRCLE_BRANCH")

    def finalize_options(self):
        pass

    def run(self):
        git_version = self.git_tag

        if git_version is None:
            branch = self.git_branch or "NO TAG/RELEASE BRANCH"
            _, _, git_version = branch.partition("release/")

        print(git_version)
        version = get_version()

        if git_version != version:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                git_version, version
            )
            sys.exit(info)


setup(
    name="mondobrain",
    version=get_version(),
    description="MondoBrain API wrapper",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="MondoBrain",
    url="https://bitbucket.org/mondobrain/mondobrain-python",
    project_urls={
        "MondoBrain homepage": "https://mondobrain.com",
        "MondoBrain source": "https://bitbucket.org/mondobrain/mondobrain-python",
    },
    packages=find_packages(exclude=["tests*"], include=["mondobrain", "mondobrain.*"]),
    package_data={"mondobrain": ["examples/*.md", "datasets/data/*.csv"]},
    include_package_data=True,
    license="MIT License",
    classifiers=[
        # How mature is this project?
        "Development Status :: 3 - Alpha",
        # Intended audience
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        # Language of project
        "Natural Language :: English",
        # License
        "License :: OSI Approved :: MIT License",
        # Versions supported
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        # Operating systems
        "Operating System :: OS Independent",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
    ],
    platforms="any",
    python_requires=">=3.6",
    install_requires=[
        "auth0-python ~= 3.13.0",
        "category_encoders ~= 2.2",
        "numpy ~= 1.18.1",
        "pandas ~= 1.0.1",
        "pyarrow ~= 0.16.0",
        "requests ~= 2.7",
        "scikit-learn ~= 0.22.1",
    ],
    # extras_require={
    # "nlp": ["solver @ git+https://bitbucket.org/mondobrain/indigo@70569bb"]
    # },
    extras_require={
        "viz": ["graphviz ~= 0.14.2", "networkx ~= 2.5", "pygraphviz ~= 1.6"]
    },
    cmdclass={"verify": VerifyVersionCommand},
)



# flake8: noqa
from .api import api_test
from .auth import auth_test
from .language import language_entities
from .meta import meta_class_start, meta_exclusion_set_start, meta_result
from .requestor import APIRequestor
from .solve import solve_result, solve_start, solve_start_file

__all__ = [
    "api_test",
    "auth_test",
    "language_entities",
    "meta_class_start",
    "meta_exclusion_set_start",
    "meta_result",
    "solve_result",
    "solve_start",
    "solve_start_file",
    "APIRequestor",
]



from collections import defaultdict
import string

import numpy as np
from sklearn.preprocessing import LabelEncoder

from mondobrain.core.series import MondoSeries
from mondobrain.utils.data import is_discrete


class DDTransformer:
    """
    -------------------------------------------------
    # A convenience class to access the default dict for
    #  transforming the values to the encoded values and back
    #Example Usage:
    edf, dd = encode(df)
    dd.original_to_encoded("column1", "valueA")
    #  >>  1
    dd.encoded_to_original("column1", 1)
    #  >> "valueA"
    __________________________________________________
    """

    def __init__(self, dataframe):
        """
        -------------------------------------------------
        :param output_default_dict: the default dictionary generated by this encoder
        __________________________________________________
        """
        def_dict = defaultdict(LabelEncoder)
        clean_dataframe = dataframe.apply(DDTransformer.clean_nulls)
        clean_dataframe.apply(lambda x: def_dict[x.name].fit_transform(x))
        def_dict["__columns__"].fit_transform(clean_dataframe.columns)
        self.set_column_encoders(clean_dataframe.columns)
        self._def_dict = def_dict

    def __repr__(self):
        return "DDTranslator for values of: " + str(list(self._def_dict.keys()))

    @staticmethod
    def _enumerate_letters(count):
        """
        Generates a list of unique capital letter strings accordingly
        eg. ['A', 'B', ... , 'AA', 'BB', 'CC', ...]

        :param df: number of unique letter strings to return
        :returns: list <str>: list of capital letter strings
        """

        letters_in_alphabet = 26

        class _unique_list(list):
            """
            _unique_list extends the standard list for this application
            and will only allow items not already in the list to be appended
            NOTE: this is *only* for this application, as this implementation
            will break if items are removed

            Adds a method, 'contains' which will check the internal set of already
            seen items and prevent appending to the list in the case it is already there
            """

            def __init__(self):
                self._seen = set()

            def append(self, item):
                if item not in self._seen:
                    self._seen.add(item)
                    super().append(item)  # append the item to the list

            def contains(self, item):
                return item in self._seen

        results = _unique_list()

        for i in range(count):
            column_title = string.ascii_uppercase[i % letters_in_alphabet]
            # must loop here in the instance that the column_title is in the set already
            while True:
                if results.contains(column_title):
                    column_title += string.ascii_uppercase[i % letters_in_alphabet]
                else:
                    break
            results.append(column_title)
        return results

    def set_column_encoders(self, columns):
        codes = self._enumerate_letters(columns.size)
        self._column_encoder = dict(enumerate(codes))
        self._column_decoder = {code: idx for idx, code in enumerate(codes)}

    def original_to_encoded_column(self, column_label):
        """
        -------------------------------------------------
        :param column_label: the column name for the lookup
        :return: returns the encoded column_label
        :raises: dd.original_to_encoded("wrong_column_name", 1) =>
             sklearn.exceptions.NotFittedError
        __________________________________________________
        """
        encoded_col_label = self._def_dict["__columns__"].transform([column_label])[0]
        encoded_col_label = self._column_encoder[encoded_col_label]

        return encoded_col_label

    def original_to_encoded(self, column_label, orig_value):
        """
        -------------------------------------------------
        :param column_label: the column name for the lookup
        :param orig_value: the original values
        :return: returns the encoded value mapped to that original value,
        for this column_label
        :raises: dd.original_to_encoded("wrong_column_name", 1) =>
             sklearn.exceptions.NotFittedError
        :raises: dd.original_to_encoded("petallengthincm", "999") =>
             ValueError if no 999 in original data set
        __________________________________________________
        """
        clean_orig_value = DDTransformer.clean_nulls(orig_value)
        encoded = self._def_dict[column_label].transform(clean_orig_value)

        if is_discrete(orig_value.dtype):
            encoded = encoded.astype(str)

        encoded_col_label = self.original_to_encoded_column(column_label)

        return encoded, encoded_col_label

    def encoded_to_original_column(self, column_label):
        """
        -------------------------------------------------
        :param column_label: the column name for the lookup
        :return: returns the original column_label
        :raises: dd.encoded_to_original("wrong_column_name", 1) =>
            sklearn.exceptions.NotFittedError
        __________________________________________________
        """
        decoded_col_label = self._def_dict["__columns__"].inverse_transform(
            [column_label]
        )[0]

        return decoded_col_label

    def encoded_to_original(self, column_label, encoded_value):
        """
        -------------------------------------------------
        :param column_label: the column name for the lookup
        :param encoded_value: the encoded values to get originals
        :return: returns the original value mapped to that encoded value,
        for this column_label
        :raises: dd.encoded_to_original("wrong_column_name", 1) =>
            sklearn.exceptions.NotFittedError
        :raises: dd.encoded_to_original("petallengthincm", "999") =>
            ValueError if no 999 in encoded data set
        __________________________________________________
        """
        decoded_col_label = self.encoded_to_original_column(column_label)

        decoded = self._def_dict[decoded_col_label].inverse_transform(
            encoded_value.astype(np.int)
        )

        return MondoSeries(decoded, name=decoded_col_label)

    def column_has_null_values(self, column):
        """
        -------------------------------------------------
        :param column_label: the column name for the lookup
        :return: returns True if variable is has null values, and False if it does not
        :raises: dd.variable_is_continuous("alskdjalskjdaldj") =>
            KeyError if no variable named 'alskdjalskjdaldj' in data set
        __________________________________________________
        """
        return column.isnull().any()

    def variable_is_continuous(self, column_label):
        """
        -------------------------------------------------
        :param column_label: the column name for the lookup
        :return: returns True if variable is continuous, and False if it is not
        :raises: dd.variable_is_continuous("alskdjalskjdaldj") =>
            KeyError if no variable named 'alskdjalskjdaldj' in data set
        __________________________________________________
        """
        return self._column_label_is_continuous[column_label]

    @staticmethod
    def clean_nulls(dataseries):
        """
        Function operates on a column and transforms nulls to a base type reference.
        This is required because null or NaN types are generally float distinct
        float values and will be encoded distinctly otherwise
        :param dataseries:
        :return:
        """
        clean_me = dataseries.copy()

        # these values will be used as replacement values
        # to ensure that they float to the top
        min_int = np.iinfo(np.int64).min
        min_float = np.finfo(np.float64).min
        min_str = "      "

        # define a dict with objects to replace nulls
        the_cleaner = (
            (np.object_, min_str),
            (np.int64, min_int),
            (np.float64, min_float),
        )

        # run pd.fillna replacements with each of the columns
        # sets and fill according to type
        for obj, base in the_cleaner:
            if np.issubdtype(clean_me.dtype, obj):
                clean_me.fillna(base, inplace=True)

        return clean_me



import os

import pandas as pd
import pytest

import mondobrain as mb


@pytest.fixture
def datapath():
    """
    Get the path to a data file.
    Parameters
    ----------
    path : str
        Path to the file, relative to ``mondobrain/tests/``
    Returns
    -------
    path including ``mondobrain/tests``.
    Raises
    ------
    ValueError
        If the path doesn't exist
    """
    BASE_PATH = os.path.join(os.path.dirname(__file__), "tests")

    def deco(*args):
        path = os.path.join(BASE_PATH, *args)
        if not os.path.exists(path):
            raise ValueError(f"Could not find file {path}")
        return path

    return deco


@pytest.fixture
def iris(datapath):
    """
    The iris dataset as a MondoDataFrame.
    """
    df = pd.read_csv(datapath("data", "iris.csv"))
    return mb.MondoDataFrame(df)


@pytest.fixture
def titanic(datapath):
    """
    The titanic dataset as a MondoDataFrame.
    """
    df = pd.read_csv(datapath("data", "titanic.csv"))
    return mb.MondoDataFrame(df)



import numpy as np
import pandas as pd

from mondobrain.solver import solve
from mondobrain.utils.data import is_continuous


def extract_rule(
    df: pd.DataFrame,
    predicted: np.ndarray,
    outcome: str = None,
    target: str = None,
    random_state=None,
    **kwargs,
):
    # Make local copy of the df
    df = df.copy()

    if outcome is None:
        outcome = df.columns[0]

    if target is None:
        target = "max" if is_continuous(df[outcome]) else "correct"

    if is_continuous(df[outcome]):
        if target not in ["max", "min"]:
            raise ValueError("target must be `min` or `max` when outcome is continuous")

        df["__error"] = np.abs(predicted - df[outcome])
    else:
        if target not in ["correct", "incorrect"]:
            raise ValueError(
                "target must be `correct` or `incorrect` when outcome is discrete"
            )

        df["__error"] = np.where((predicted == df[outcome]), "correct", "incorrect")

    solve_df = df.drop(outcome, axis=1)
    return solve(
        solve_df, outcome="__error", target=target, random_state=random_state, **kwargs
    )



import numpy as np

import mondobrain as mb

try:
    from indigo.camp import Tent
except Exception:
    raise Exception(
        "NLP tools have not been installed. Try running "
        "`pip install git+https://bitbucket.org/mondobrain/indigo@master#egg=indigo`"
    )


class MondoBot(object):
    """
    builds an object to ask a natural language question/demand
    and get a natural language answer back
    """

    def __init__(self, frame, message=None):
        self.solver = mb.Solver()

        self.frame = frame.copy()
        self.frame.columns = self.frame.columns.str.lower()
        self.message = message
        self.tent = Tent(dataset=self.frame, question=message)

        self._operation_dict = {
            "min": "minimized",
            "max": "maximized",
        }

    def __repr__(self):
        _repr = list()

        _repr.append(str(self.__class__))
        for feat in self.tent.stakes["features"]:

            explorable = outcome = constraint = "N/A"
            if "entities" in self.tent.stakes:
                explorable, outcome, constraint = "No", "No", "None"
                if feat["key"].lower() in (
                    expl.lower() for expl in self.tent.stakes["entities"]["explorable"]
                ):
                    explorable = "Yes"
                if (
                    feat["key"].lower()
                    == self.tent.stakes["entities"]["outcome"].lower()
                ):
                    if (
                        self.tent.stakes["entities"]["operation"]["type"]
                        == "continuous"
                    ):
                        outcome = f"Yes; {self._operation_dict[self.tent.stakes['entities']['operation']['value']]}"  # NOQA E501
                    else:
                        outcome = f"Yes; {feat['key']} = {self.tent.stakes['entities']['operation']['value']}"  # NOQA E501
                for const in self.tent.stakes["entities"]["constraints"]:
                    if feat["key"].lower() == const["feature"].lower():
                        constraint = " ".join(const.values())

            if feat["type"] == "continuous":
                _repr.append(
                    f"variable: {feat['key']}\n"
                    f"\t-type: {feat['type']}\n"
                    f"\t-range: {feat['range'][0]} to {feat['range'][1]}\n"
                    f"\t-explorable: {explorable}\n"
                    f"\t-outcome: {outcome}\n"
                    f"\t-constraint: {constraint}\n"
                )
            else:
                _repr.append(
                    f"variable: {feat['key']}\n"
                    f"\t-type: {feat['type']}\n"
                    f"\t-modalities: {', '.join(tuple(feat['modalities'][i] for i in range(5) if i < len(feat['modalities'])))}{'...' if len(feat['modalities']) > 5 else ''}\n"  # NOQA E501
                    f"\t-explorable: {explorable}\n"
                    f"\t-outcome: {outcome}\n"
                    f"\t-constraint: {constraint}\n"
                )

        return "\n".join(_repr)

    def _distinguish_discrete(self, rule_values):
        if isinstance(rule_values, str):
            return rule_values
        else:
            return f"between {rule_values['lo']} and {rule_values['hi']}"

    def _process_constraint(self, col, constraint, constraint_class=None):
        constraints = constraint.split(" ")
        if "=" in constraints:
            return self.frame[col] == constraint_class
        elif ">" in constraints:
            bound = float(constraints[1])
            return self.frame[col] > bound
        elif "<" in constraints:
            bound = float(constraints[1])
            return self.frame[col] < bound
        elif ">=" in constraints:
            bound = float(constraints[1])
            return self.frame[col] >= bound
        elif "<=" in constraints:
            bound = float(constraints[1])
            return self.frame[col] <= bound
        elif "between" in constraints:
            low, high = float(constraints[1]), float(constraints[3])
            return (self.frame[col] > low) & (self.frame[col] < high)
        else:
            return None

    def _verbalize(self, entities, rule):
        discrete_outcome = entities["operation"]["type"] == "discrete"
        if discrete_outcome:
            outcome_statement = (
                f"most significant with respect to \"{entities['operation']['value']}\""
            )
        else:
            outcome_statement = self._operation_dict[entities["operation"]["value"]]
        answer = f"\"{entities['outcome']}\" is {outcome_statement} when:\n"
        for column, result in rule.items():
            between = self._distinguish_discrete(result)
            answer += f'\t- "{column}" is {between}\n'
        if entities["constraints"]:
            answer += "\nconstraints applied:\n"
            for constraint in entities["constraints"]:
                answer += f"- \"{' '.join(constraint.values())}\"\n"

        return answer

    def ask(self, message=None):

        if message is None:
            message = self.message

        self.tent.pitch(question=message)
        entities = self.tent.stakes["entities"]
        self._repr = dict()

        mask = np.ones(self.frame.shape[0], dtype=bool)
        if entities["constraints"]:
            for constraint in entities["constraints"]:
                if "class" not in constraint:
                    col, equality = constraint["feature"], constraint["constraint"]
                    mask = np.bitwise_and(
                        mask, self._process_constraint(col.lower(), equality)
                    )
                else:
                    col, equality, constraint_class = (
                        constraint["feature"],
                        constraint["constraint"],
                        constraint["class"],
                    )
                    mask = np.bitwise_and(
                        mask,
                        self._process_constraint(
                            col.lower(), equality, constraint_class
                        ),
                    )

        mb_X = mb.MondoDataFrame(
            self.frame.loc[mask, [col.lower() for col in entities["explorable"]]]
        )
        mb_y = mb.MondoSeries(
            self.frame.loc[mask, entities["outcome"].lower()], name=entities["outcome"]
        )
        mb_y.target_class = entities["operation"]["value"]

        self.solver.fit(mb_X, mb_y)

        print(self._verbalize(entities=entities, rule=self.solver.rule))



# flake8: noqa
import logging

__author__ = "MondoBrain"
__version__ = "0.1.3"

# Configuration variables
api_key = None
api_secret = None
api_token = None

api_base = "https://api.prod.mondobrain.com/"
auth0_domain = "mb-production.us.auth0.com"

verify_ssl_certs = True
proxy = None
default_http_client = None
enable_telemetry = True
max_network_retries = 0
ca_bundle_path = None

# Set to either 'debug' or 'info', controls console logging
log = None

from mondobrain import api  # isort:skip
from mondobrain.core.api import MondoDataFrame, MondoSeries  # isort:skip
from mondobrain.prescriber import Solver  # isort:skip



from pandas import DataFrame

from mondobrain.core.series import MondoSeries


class MondoDataFrame(DataFrame):
    """
    A MondoDataFrame object is a pandas.DataFrame that has certain
    MondoBrain specific operations and methods.

    The MondoDataFrame object is composed of 1 or more MondoSeries
    objects. The inner workings of the MondoDataFrame ensures that
    each variable is instantiated as a MondoSeries.

    For additional reference specific to the Pandas DataFrame object
    and features, please refer to the Pandas API

    **Parameters**

    data: ndarray, Iterable, dict, DataFrame, or MondoDataFrame
        Dict can contain Series, MondoSeries, arrays, constants, or list-like objects.

    **Examples**

    >>> from mondobrain.MondoDataframe import MondoDataFrame
    >>> import Pandas as pd
    >>> df = pd.read_csv('path/to/file.csv')
    >>> mdf = MondoDataFrame(df)

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return MondoDataFrame

    @property
    def _constructor_sliced(self):
        return MondoSeries



import logging
import os
import re
import sys

import mondobrain

MONDOBRAIN_LOG = os.environ.get("MONDOBRAIN_LOG")

logger = logging.getLogger("mondobrain")


def _console_log_level():
    if mondobrain.log in ["debug", "info"]:
        return mondobrain.log
    elif MONDOBRAIN_LOG in ["debug", "info"]:
        return MONDOBRAIN_LOG
    else:
        return None


def log_debug(message, **params):
    msg = logfmt(message, dict(**params))
    if _console_log_level() == "debug":
        print(msg, file=sys.stderr)
    logger.debug(msg)


def log_info(message, **params):
    msg = logfmt(message, dict(**params))
    if _console_log_level() in ["debug", "info"]:
        print(msg, file=sys.stderr)
    logger.info(msg)


def logfmt(message, props: dict):
    def fmt(key: str, val):
        if hasattr(val, "decode"):
            val = val.decode("utf-8")

        if not isinstance(val, str):
            val = str(val)

        if re.search(r"\s", val):
            val = repr(val)

        if re.search(r"\s", key):
            key = repr(key)

        return "{key}={val}".format(key=key, val=val)

    prop_str = " ".join([fmt(key, val) for key, val in sorted(props.items())])

    return "%s %s" % (message, prop_str)
